from vkconnections import main as m
import sys

class Create:
	# keys - list of given keys
	def __init__(self, keys):
		if len(keys) != 4:
			sys.exit("List of keys should have 4 different keys.")
		self.keys = keys
		
	def getConnection(self, userFrom, userTo):
		return m.getWayBetween(userFrom, userTo, self.keys)

	def printConnection(result, photo=False):
		for way in temp:
			for person in way:
				s = "Id: " + str(person['id']) + " " + "Name: " + person['full_name']
				s += "Photo: " + person['photo'] if photo else ""
				print(s)
			print()


from vkconnections import VkAPI as API
from vkconnections import SecondLevel as division, Helper as h
import threading

firstPair, secondPair = {1: "", 2: ""}, {1: "", 2: ""}

class FirstDivision:
    def __init__(self, twoTokens, friends, threadName):
        self.twoTokens = twoTokens
        self.friends = friends
        self.threadName = threadName
        self.data = {}

    def makeCalculations(self):
        self.data = division.getListOfFriends(self.friends, self.twoTokens, self.threadName)

    def getData(self):
        return self.data


def calculations(fromId, friendsUserFrom, toId, friendsUserTo, keys):
    global firstPair
    global secondPair

    firstPair = {1: keys[0], 2: keys[1]}
    secondPair = {1: keys[2], 2: keys[3]}

    friendsUserFrom, friendsUserTo = h.checkForNormalLength(friendsUserFrom, friendsUserTo)
    firstHalfOfFriends, secondHalfOfFriends = getHalfOfFriends(friendsUserFrom, friendsUserTo)

    adjFrom, adjTo = makeFirstDivision(firstHalfOfFriends, secondHalfOfFriends)

    circles = getCircles(friendsUserFrom, friendsUserTo, adjFrom, adjTo, fromId, toId)

    allTheWays = getWays(fromId, toId, circles)

    outputList = API.getExecute(API.getVkScriptForUsersExecute(allTheWays), firstPair[1])
    resultList = h.makeListsNicer(outputList)
    return resultList


def getHalfOfFriends(friendsUserFrom, friendsUserTo):
    allTheFriends = friendsUserFrom + friendsUserTo
    middle = len(allTheFriends) // 2
    firstHalfOfFriends = allTheFriends[:middle]
    secondHalfOfFriends = allTheFriends[middle:]
    return firstHalfOfFriends, secondHalfOfFriends


def makeFirstDivision(firstHalfOfFriends, secondHalfOfFriends):
    firstHalfFirstDivision = FirstDivision(firstPair, firstHalfOfFriends, 'First half: ')
    secondHalfFirstDivision = FirstDivision(secondPair, secondHalfOfFriends, 'Second half: ')

    adjFrom = threading.Thread(target=firstHalfFirstDivision.makeCalculations, name='FirstDivision - 1')
    adjTo = threading.Thread(target=secondHalfFirstDivision.makeCalculations, name='FirstDivision - 2')

    adjFrom.start()
    adjTo.start()

    adjFrom.join()
    adjTo.join()

    # now adj is a dictionary where key - id of a person, value - ids of his friends
    adjFrom = firstHalfFirstDivision.getData()
    adjTo = secondHalfFirstDivision.getData()
    return adjFrom, adjTo


def getCircles(friendsUserFrom, friendsUserTo, adjFrom, adjTo, fromId, toId):
    circles = {}
    allFirstFriends = friendsUserFrom + friendsUserTo  # temp
    allAdj = adjFrom + adjTo

    for index in range(len(allAdj)):
        circles[allFirstFriends[index]] = allAdj[index]

    circles = h.clearCircles(circles)
    circles = h.getItems(circles)

    circles[fromId] = friendsUserFrom
    circles[toId] = friendsUserTo
    circles = h.fillAllTheVertices(circles)
    return circles


def getWays(fromId, toId, circles):
    allTheWays = []
    count = 1
    parents = h.BFS(s=fromId, Adj=circles)
    way = h.algorithm(parents, toId)
    allTheWays.append(way)

    while way and count < 3:
        count += 1
        way = way[1:-1]
        for item in way:
            circles.pop(item)
        parents = h.BFS(s=fromId, Adj=circles)
        way = h.algorithm(parents, toId)
        allTheWays.append(way)
    return allTheWays



import random

def getPureLink(link):
    link = link.strip().lower()
    if 'vk.com/' in link:
        link = link.split('vk.com/')[1]
    if link.startswith('/'):
        link = link.split('/')[1]
    if link.endswith('/'):
        link = link.split('/')[0]
    if link.startswith('id'):
        link = link[2:]
    return link

def checkForNormalLength(friendsUserFrom, friendsUserTo):
    if len(friendsUserFrom) + len(friendsUserTo) > 1500:
        if len(friendsUserFrom) > len(friendsUserTo):
            friendsUserFrom = listCut(friendsUserFrom)
            friendsUserFrom, friendsUserTo = checkForNormalLength(friendsUserFrom, friendsUserTo)
        else:
            friendsUserTo = listCut(friendsUserTo)
            friendsUserFrom, friendsUserTo = checkForNormalLength(friendsUserFrom, friendsUserTo)
    return friendsUserFrom, friendsUserTo

def listCut(friendsList):
    random.shuffle(friendsList)
    cut = len(friendsList) // 2
    return friendsList[:cut]


def clearCircles(circles):
    circles = {k: v for k, v in circles.items() if v is not None}
    circles = {k: v for k, v in circles.items() if v is not False}
    return circles


def getItems(circles):
    returnDict = {}
    for key in circles:
        returnDict[key] = circles[key].get('items')
    return returnDict


def fillAllTheVertices(circles):
    tempDict = {}
    for item in circles:
        for index in circles[item]:
            if index not in circles:
                if tempDict.get(index):
                    tempDict[index].append(item)
                else:
                    tempDict[index] = [item]
    for item in tempDict:
        circles[item] = tempDict[item]
    return circles

def makeListsNicer(outputList):
    resultList = []
    tempList = []
    for index in outputList:
        for item in index:
            tempList.append(
                {'id': item['id'],
                 'full_name': item['first_name'] + ' ' + item['last_name'],
                 'photo': item['photo_100']})
        resultList.append(tempList)
        tempList = []
    return resultList

def BFS(s, Adj):
    parent = {s: None}
    frontier = [s]
    while frontier:
        next = []
        for u in frontier:
            if u in Adj:
                for v in Adj[u]:
                    if v not in parent:
                        parent[v] = u
                        next.append(v)
        frontier = next
    return parent


def getWay(parents, listFinal, userTo):
    for item in parents:
        if item == userTo:
            listFinal.append(parents[item])
            getWay(parents, listFinal, parents[item])
    return listFinal


def algorithm(parents, userTo):
    listFinal = []
    listFinal = getWay(parents, listFinal, userTo)
    if listFinal:
        listFinal.remove(None)
        listFinal.reverse()
        listFinal.append(userTo)
        return listFinal
    else:
        return []


def getWayCount(output):
    if isinstance(output, list):
        return [len(item) for item in output]
    return None




from vkconnections import VkAPI as API, APIKeys as k
from vkconnections import FirstLevel as algo


# works for get way between users only
def getWayBetween(fromId, toId, keys):
    API.setKeys(keys)
    users = API.getUsers(fromId + ',' + toId)
    if users:
        fromId = users[0].get('id')  # id userFrom
        toId = users[1].get('id')  # id of userTo

        code = API.getVkScriptForExecute([fromId, toId])[0]
        friends = API.getExecute(code, keys[1])
        friendsUserFrom = friends[0].get('items')  # list of ids of friends of userFrom
        friendsUserTo = friends[1].get('items', False)  # list of ids of friends of userTo
        if not friendsUserFrom and friendsUserTo:
            return k.noSuchUser
        if len(friendsUserFrom) == 0 or len(friendsUserTo) == 0:
            return k.noFriends
        result = algo.calculations(fromId, friendsUserFrom, toId, friendsUserTo, keys)
        if len(result) == 0:
            return k.tooFar
        return result
    return k.noSuchUser


firstPair = {1: "", 2: ""}
secondPair = {1: "", 2: ""}

def setKeys(keys):
	global firstPair
	global secondPair
	firstPair = {1: keys[0], 2: keys[1]}
	secondPair = {1: keys[2], 2: keys[3]}


# Vk API Constants
api_url = 'https://api.vk.com/method/'
method_name_friends = 'friends.get?'
method_name_user = 'users.get?'
method_name_execute = 'execute?'
access_token = 'access_token='
version = 'v=5.92'

# users.get
usersGet = api_url + method_name_user + 'user_ids=' + '{0}' + '&fields=photo_100' + '&' + access_token + '{1}' + '&' + version

# friends.get
friendsGet = api_url + method_name_friends + 'user_id=' + '{0}' + '&' + access_token + '{1}' + '&' + version

# execute
executeGet = api_url + method_name_execute + 'code={}&' + access_token + '{}' + '&' + version

# errors
noSuchUser = {'error': {'code': '001', 'message': 'The user is not found'}}
noFriends = {'error': {'code': '002', 'message': 'The user either does not have or has hidden friends'}}
tooFar = {'error': {'code': '002', 'message': 'The distance between two users is too far'}}


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="vkconnections",
    version="0.0.6",
    author="Andrey Volkov",
    author_email="andrey@volkov.tech",
    description="Finding the shortest path from one vk user to another through friends.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AndreyVolkovBI/VkConnections",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


import requests
import math
import time
from json import loads
from vkconnections import APIKeys as c

keys = {}

def setKeys(vkKeys):
    global keys
    keys = vkKeys


def getUsers(usersUrl):
    global keys
    temp = None
    count = 0
    while temp is None:
        response = requests.get(c.usersGet.format(str(usersUrl), keys[0])).text
        response = loads(response)
        temp = response.get('response')
        if count > 1:
            time.sleep(1)
        count += 1
        if count == 50:
            return None
    return response.get('response')


def getFriends(id):
    temp = None
    count = 0
    while temp is None:
        response = requests.get(c.friendsGet.format(str(id), keys[0])).text
        response = loads(response)
        temp = response.get('response')
        if count > 1:
            time.sleep(1)
        count += 1
        if count == 50:
            return None
    return response.get('response').get('items')


def getExecute(code, token):
    temp = None
    count = 0
    while temp is None:
        myString = c.executeGet.format(code, token)
        execute = requests.get(myString).text
        execute = loads(execute)
        temp = execute.get('response')
        if count > 1:
            time.sleep(1)
        count += 1
        if count == 50:
            return None
    return execute.get('response')


def getVkScriptForExecute(allTheFriends):
    lenFriends = len(allTheFriends)
    requestsCount = lenFriends // 25 + math.ceil((lenFriends - lenFriends // 25 * 25) / 25)
    outputString = []
    startIndex = 0
    endIndex = 25
    for number in range(0, requestsCount):
        requestString = 'return [ '
        for item in range(startIndex, endIndex):
            if lenFriends > item:
                requestString += 'API.friends.get({"user_id": ' + str(allTheFriends[item]) + ' }), '
        requestString = requestString[0:-2]
        requestString += ' ];'
        outputString.append(requestString)
        startIndex += 25
        endIndex += 25
    return outputString

def getVkScriptForUsersExecute(allTheUsers):
    requestString = 'return [ '
    for index in range(len(allTheUsers)):
        stringOfUsers = ""
        for item in allTheUsers[index]:
            stringOfUsers += str(item) + ','
        stringOfUsers = stringOfUsers[0:-1]
        requestString += 'API.users.get({"user_ids": "' + stringOfUsers + '", "fields": "photo_100" }), '
    requestString = requestString[0:-1]
    requestString += ' ];'
    return requestString



from vkconnections import VkAPI as API
import threading


class SecondDivision:
    def __init__(self, listOfVkScript, token):
        self.listOfVkScript = listOfVkScript
        self.token = token
        self.data = []

    def execute(self):
        listOutput = []
        for item in self.listOfVkScript:
            temp = API.getExecute(item, self.token)
            listOutput.append(temp)
        self.data = listOutput

    def getData(self):
        return self.data


# input: list of users' ids
# output: list of lists of friends
def getListOfFriends(allTheUsers, twoTokens, threadName):
    firstList, secondList = getHalfOfVkScript(allTheUsers)
    firstList, secondList = makeSecondDivision(firstList, secondList, twoTokens, threadName)
    return getFullList(firstList, secondList)


def getHalfOfVkScript(allTheUsers):
    listOfVkScript = API.getVkScriptForExecute(allTheUsers)

    middle = len(listOfVkScript) // 2
    firstList = listOfVkScript[:middle]
    secondList = listOfVkScript[middle:]
    return firstList, secondList


def makeSecondDivision(firstList, secondList, twoTokens, threadName):
    firstHalfSecondDivision = SecondDivision(firstList, twoTokens[1])
    secondHalfSecondDivision = SecondDivision(secondList, twoTokens[2])

    firstList = threading.Thread(target=firstHalfSecondDivision.execute, name=threadName + 'Second Division - 1')
    secondList = threading.Thread(target=secondHalfSecondDivision.execute, name=threadName + 'Second Division - 2')

    firstList.start()
    secondList.start()

    firstList.join()
    secondList.join()

    firstList = firstHalfSecondDivision.getData()
    secondList = secondHalfSecondDivision.getData()
    return firstList, secondList


def getFullList(firstList, secondList):
    listOutput = firstList + secondList
    returnList = []
    for item in listOutput:
        for iterator in item:
            returnList.append(iterator)
    return returnList
